#========================================================================================
# package      DE_EPAGES::Order::API::Object::Basket
# state        public
# base         DE_EPAGES::Shop::API::Object::ShopObject
# description  object interface for Basket.
#========================================================================================
package DE_EPAGES::Order::API::Object::Basket;
use base DE_EPAGES::Shop::API::Object::ShopObject;

use strict;
use Scalar::Util qw ( weaken );
use DE_EPAGES::Core::API::PerlTools qw ( Compare ElementOf );
use DE_EPAGES::Customer::API::Shop qw ( GetCustomerForUser );
use DE_EPAGES::Database::API::Connection qw( GetCurrentDBHandle );
use DE_EPAGES::DataCache::API::Cache qw( UpdateLocalCache );
use DE_EPAGES::Object::API::Factory qw ( LoadClassByAlias LoadObject Transaction );
use DE_EPAGES::Order::API::LineItemContainer qw (SaveShipping SavePayment);
use DE_EPAGES::Shop::API::Address qw ( GetTaxCountryIDForShop IsEUAddress );
use DE_EPAGES::Tax::API::Shop qw( TaxMatrixByEUAddress );
use DE_EPAGES::Trigger::API::Trigger qw ( TriggerHook );

#========================================================================================
# function     className
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ClassName = $Basket->className;
#----------------------------------------------------------------------------------------
# description  Returns the class name, used for DAL access.
#----------------------------------------------------------------------------------------
# return       $ClassName | class name | String
#========================================================================================
sub className  { 'Basket' }

#========================================================================================
# function     dbPackage
# state        public
#----------------------------------------------------------------------------------------
# syntax       $DbPackage = $Basket->dbPackage;
#----------------------------------------------------------------------------------------
# description  Returns the database package, used for DAL access.
#----------------------------------------------------------------------------------------
# return       $DbPackage | database package name | String
#========================================================================================
sub dbPackage { __PACKAGE__ }

#========================================================================================
# function     insert
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Basket = DE_EPAGES::Order::API::Object::Basket->insert($hInfo);
# example      my $Class = LoadClassByAlias('Basket');
#               my $Basket = $Class->insertObject($hInfo);
#----------------------------------------------------------------------------------------
# description  Creates a new basket (creates also LineItemContainer, ShippingLineItem, PaymentLineItem).
#               Afterwards the hook 'OBJ_InsertBasket' is triggered.
#----------------------------------------------------------------------------------------
# input        $hInfo | attributes of new object, required keys:
#               <ul>
#                 <li>Shop - shop - object
#                 <li>User - user - object
#               </ul> | ref.hash
# return       $Basket | new object | object
# hook         OBJ_InsertBasket | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>BasketID      - object identifier    - int
#               </ul>
#========================================================================================
sub insert {
    my ($class, $hInfo) = @_;

    $hInfo->{'User'} //= $hInfo->{'Parent'}->get('Parent');

    my $Shop = $hInfo->{'Shop'} // $hInfo->{'User'}->getSite();
    $hInfo->{'Alias'} //= $Shop->id;

    # test if we can create a basket
    $Shop->featureCurrentValue('Basket',1);

    my $hBasketAttributes = {%$hInfo};
    my $Class = LoadClassByAlias($class->className);
    delete $hBasketAttributes->{$_} foreach grep { ! $Class->existsAttribute($_) } keys %$hBasketAttributes;

    my $self = $class->SUPER::insert($hBasketAttributes);

    unless( $hInfo->{'OrderGUID'} ) {
        # insert container unless the basket is imported via XML
        my %ContainerInfo;
        $ContainerInfo{'Alias'} = 'LineItemContainer';
        $ContainerInfo{$_} = $hInfo->{$_} foreach grep { defined $hInfo->{$_} } qw ( LocaleID TaxArea CurrencyID TaxModel );
        if (defined $hInfo->{'User'}) {
            $ContainerInfo{'LocaleID'}   = $hInfo->{'User'}->get('LocaleID')   unless defined $ContainerInfo{'LocaleID'};
            $ContainerInfo{'CurrencyID'} = $hInfo->{'User'}->get('CurrencyID') unless defined $ContainerInfo{'CurrencyID'};
        }
        if (defined $hInfo->{'Customer'}) {
            $ContainerInfo{'TaxModel'}   = $hInfo->{'Customer'}->get('TaxModel') unless defined $ContainerInfo{'TaxModel'};
        }
        if (defined $Shop) {
            $ContainerInfo{'TaxModel'}   = $Shop->get('TaxModel')   unless defined $ContainerInfo{'TaxModel'};
            $ContainerInfo{'LocaleID'}   = $Shop->get('LocaleID')   unless defined $ContainerInfo{'LocaleID'};
        }
        $ContainerInfo{'LastPaymentMethod'}  = $hInfo->{'PaymentMethod'}  if defined $hInfo->{'PaymentMethod'};
        $ContainerInfo{'LastShippingMethod'} = $hInfo->{'ShippingMethod'} if defined $hInfo->{'ShippingMethod'};

        my $Container = $self->insertChild('LineItemContainer', \%ContainerInfo);
        $hBasketAttributes->{'LineItemContainer'} = $Container;
    }

    $self->set($hBasketAttributes);


    # test if we have a TaxMethod
    if($Shop->get('TaxMethod')){
        $Shop->get('TaxMethod')->updateContainerTaxArea($self->container);
    }else {
        $self->updateTaxArea();
        $self->updateTaxAreaDigital();
    }

    %$hInfo = (); # dont need to set after insert

    TriggerHook('OBJ_InsertBasket', {'Object' => $self, 'BasketID' => $self->id });
    return $self;
}

#========================================================================================
# function     delete
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Basket->delete;
#----------------------------------------------------------------------------------------
# description  Deletes this object, therefore the hook 'OBJ_DeleteBasket' is triggered.
#----------------------------------------------------------------------------------------
# hook         OBJ_DeleteBasket | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>BasketID      - object identifier    - int
#               </ul>
#========================================================================================
sub delete {
    my $self = shift;

    # EPG-13756: delete the local cache of child objects in case another appserver
    # has already created an order for this basket. In this case we might delete the
    # LineItemContainer which was a child object of the basket, but isn't anymore.
    UpdateLocalCache(GetCurrentDBHandle()->cacheKey, 'ObjectByParentID', $self->id);

    TriggerHook('OBJ_DeleteBasket', {'Object' => $self, 'BasketID' => $self->id });
    $self->set({
        'Customer' => undef,
        'LineItemContainer' => undef,
        'BillingAddress' => undef,
        'ShippingAddress' => undef,
        'ShopAddress' => undef,
    });
    $self->SUPER::delete;
    return;
}

#========================================================================================
# function     set
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Basket->set($hValues, $LanguageID);
#----------------------------------------------------------------------------------------
# description  This function triggers hooks 'OBJ_BeforeUpdateBasket' and 'OBJ_AfterUpdateBasket'.
#----------------------------------------------------------------------------------------
# input        $hValues | attribute name => value hash | ref.hash
# input        $LanguageID | (optional; required for localized attributes) language id | int
#----------------------------------------------------------------------------------------
# hook         OBJ_BeforeUpdateBasket | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>BasketID      - object identifier    - int
#                 <li>Values        - new values for object - ref.hash
#                 <li>LanguageID    - values of this language will be updated - int
#               </ul>
# hook         OBJ_AfterUpdateBasket | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>BasketID      - object identifier    - int
#                 <li>Values        - updated values       - ref.hash
#                 <li>LanguageID    - values updated of this language - int
#               </ul>
#========================================================================================
sub set {
    my $self = shift;
    my ($hValues, $LanguageID) = @_;

    TriggerHook('OBJ_BeforeUpdateBasket', {'Object'=>$self, 'Values'=> $hValues, 'BasketID' => $self->id, 'LanguageID' => $LanguageID});
    if (exists $hValues->{'User'} and not exists $hValues->{'Customer'}) {
        my $Shop = exists $hValues->{'Shop'} ? $hValues->{'Shop'} : $self->getSite();
        my $Customer = GetCustomerForUser($hValues->{'User'}, $Shop);
        my $OldCustomer = $self->get('Customer');
        if (Compare('Customer', $OldCustomer, $Customer)) {
            $hValues->{'Customer'} = $Customer;
        }
    }
    if (exists $hValues->{'ShippingAddress'} or exists $hValues->{'BillingAddress'}) {
        $hValues->{'IsAddressOK'} = 0;
    }
    $self->SUPER::set($hValues, $LanguageID);
    TriggerHook('OBJ_AfterUpdateBasket',  {'Object'=>$self, 'Values'=> $hValues, 'BasketID' => $self->id, 'LanguageID' => $LanguageID});
    return;
}

#========================================================================================
# function     container
# state        public
#----------------------------------------------------------------------------------------
# syntax       $LineItemContainer = $Basket->container;
#----------------------------------------------------------------------------------------
# description  Returns lineitem container of basket.
#----------------------------------------------------------------------------------------
# return       $LineItemContainer | a lineitemcontainer object | object
#========================================================================================
sub container {
    my $self = shift;

    return undef unless defined $self->{'Attributes'}{'LineItemContainerID'};

    my $LineItemContainer = $self->{'LineItemContainer'};
    unless( defined $LineItemContainer ) {
        $LineItemContainer = LoadObject( $self->{'Attributes'}{'LineItemContainerID'} );
        $self->{'LineItemContainer'} = $LineItemContainer;
        weaken $self->{'LineItemContainer'};
    }
    return $LineItemContainer;
}

#========================================================================================
# function     recalculate
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Basket->recalculate;
#----------------------------------------------------------------------------------------
# description  Recalculate lineitem container,
#               see <function DE_EPAGES::Order::API::Object::LineItemContainer::recalculate>.
#========================================================================================
sub recalculate {
    my $self = shift;

    Transaction( sub {
        $self->container->recalculate;
    });
}

#========================================================================================
# function     insertLineItem
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Basket->insertLineItem;
#----------------------------------------------------------------------------------------
# description  Insert a lineItem in to container, see <function DE_EPAGES::Order::API::Object::LineItemContainer::insertLineItem>.
#========================================================================================
sub insertLineItem    { shift->container->insertLineItem(@_)    } ## no critic qw(RequireArgUnpacking)

#========================================================================================
# function     default
# state        private
#----------------------------------------------------------------------------------------
# syntax       $hValues = DE_EPAGES::Order::API::Object::Basket->default( $Parent, $hInfo );
# example      $hDefaults = $Class->defaultAttributes($Parent, $hInfo); # for insert a new object
#               $hDefaults = $Object->class->defaultAttributes($Object->parent, $hValues); # for an existing object
#----------------------------------------------------------------------------------------
# description  Returns default attribute values for new objects.
#               Used by <function DE_EPAGES::Object::Object::Class::defaultAttributes>
#----------------------------------------------------------------------------------------
# input        $Parent | parent object of the new object | object
# input        $hInfo | attribute name => value hash (passed to insert()) | ref.hash
# return       $hValues | hash with default attribute values (name => value) | ref.hash
#========================================================================================
sub default {
    my $class = shift;
    my ($Parent, $hInfo) = @_;

    my $Customer;
    if ( exists $hInfo->{'Customer'}) {
        $Customer = $hInfo->{'Customer'};
    } else {
        $Customer = defined $hInfo->{'User'} ? GetCustomerForUser($hInfo->{'User'}, $hInfo->{'Shop'}): undef;
    }

    return {
        %{$class->SUPER::default($Parent, $hInfo)},
        'Customer' => $Customer,
        'ShopAddress' => defined $hInfo->{'Shop'} ? $hInfo->{'Shop'}->get('Address') : undef,
        'IsAddressOK' => 0,
        'IsTacOK' => 0,
        'IsCustomerInformationOK' => 0
    };
}

#========================================================================================
# function     updateTaxArea
# state        public
#----------------------------------------------------------------------------------------
# syntax       $TaxChanged = $Basket->updateTaxArea();
#----------------------------------------------------------------------------------------
# description  Sets the basket tax area based on the shipping address country
#
#               Returns true, if calculation of tax has changed, because of changed
#               country in billing or shipping address
#----------------------------------------------------------------------------------------
# return       $TaxChanged | true if tax area or tax calculation has changed,
#                             otherwise undef | boolean
#========================================================================================
sub updateTaxArea {
    my $self = shift;
    my $Shop = $self->getSite();
    my $Container = $self->container;
    return unless defined $Container;
    if (defined $Container->get('TaxArea') && $Shop->get('CanSelectTaxAreaOnBasket')) {
        $self->recalculate;
        return;
    }

    my $Customer = $self->get('Customer');
    my $TaxArea = defined $Customer ? $Customer->get('TaxArea') : undef;

    # set TaxMatrix to use if we have an other EU country
    my $hShopInfo = $Shop->get([qw(IsUnityRegistered DeliveryThresholdIsActive Address)]);
    my $ShopAddressIsEU = IsEUAddress($hShopInfo->{'Address'});
    # get shipping Country
    my $ShopCountryID = GetTaxCountryIDForShop($Shop);
    my $ShippingAddress = $self->get('ShippingAddress') // $self->get('BillingAddress');

    # use users shipping address if we have no basket shipping address
    unless (defined $ShippingAddress) {
        my $User = $self->get('User');
        if(defined $User){
            $ShippingAddress = $User->get('ShippingAddress') // $User->get('BillingAddress');
        }
    }

    my $ShippingAddressIsEU = defined $ShippingAddress ? IsEUAddress($ShippingAddress) : 0;
    my $ShippingCountryID = $Container->get('ShippingCountryID');
    $ShippingCountryID = $ShopCountryID  unless $ShippingCountryID;

    my $TaxMatrixForDeliveryThreshold = undef;
    # test if we have an active delivery threshold and ...
    if( $hShopInfo->{'DeliveryThresholdIsActive'} and $ShopAddressIsEU and $ShippingAddressIsEU ) {
        $TaxMatrixForDeliveryThreshold = TaxMatrixByEUAddress($ShippingAddress);
    }

    # now we have a TaxMatrix for the delivery address and can change the customers tax area
    # if we have a child with the same alias
    if( (defined $TaxMatrixForDeliveryThreshold) and (defined $TaxArea) ){
        my $CustomerTaxAreaAlias = $TaxArea->alias();
        $TaxArea = $TaxMatrixForDeliveryThreshold->child($CustomerTaxAreaAlias) if $TaxMatrixForDeliveryThreshold->existsChild($CustomerTaxAreaAlias);
    }

    unless (defined $TaxArea) {
        # use the correct TaxMatrix
        my $TaxMatrix = defined $TaxMatrixForDeliveryThreshold ? $TaxMatrixForDeliveryThreshold : $Shop->get('TaxMatrix');
        #zipcode
        my $Zipcode;
        $Zipcode = $ShippingAddress->get('Zipcode') if defined $ShippingAddress;
        # get TaxArea
        $TaxArea = $ShippingCountryID
                    ? $TaxMatrix->taxAreaByCountry($ShippingCountryID, $Zipcode)
                    : $TaxMatrix->fallbackTaxArea();
    }

    # set TaxArea and NoTax
    my $OldTaxArea = $Container->get('TaxArea');
    if (($TaxArea && Compare('TaxArea', $OldTaxArea, $TaxArea)) ||
        ($Container->get('NoTax') && $Container->get('TotalTax')) ||
        (!$Container->get('NoTax') && !$Container->get('TotalTax'))) {
        $Container->set({'TaxArea' => $TaxArea}) if defined $TaxArea;

        $self->_correctTaxClassesOnContainer($Container, $TaxMatrixForDeliveryThreshold) if defined $TaxMatrixForDeliveryThreshold;

        $self->recalculate;
        my $aTaxes = $Container->get('Taxes');
        foreach my $LineItemTax (@{$aTaxes}) {
            return 1  if $LineItemTax->get('LineItemPrice');
        }
    }
    return;
}
sub _correctTaxClassesOnContainer {
    my $self = shift;
    my ($Container, $NewTaxMatrix) = @_;

    my $LineItemClass = LoadClassByAlias('LineItem');
    my @LineItems = grep { $_->instanceOf($LineItemClass) } @{$self->container->children};

    foreach my $LineItem (@LineItems) {
        #delete Tax items. The lineitems will be added later by using recalculate
        if ($LineItem->instanceOf('LineItemTax')){
            $LineItem->delete();
            next;
        }

        my $OldTaxClass = $LineItem->get('TaxClass');
        next unless defined $OldTaxClass;

        my $NewTaxClass = $NewTaxMatrix->getTaxClassByAlias($OldTaxClass->alias) ;
        $LineItem->set({'TaxClass' => $NewTaxClass } );

    }
}
#========================================================================================
# function     updateTaxAreaDigital
# state        public
#----------------------------------------------------------------------------------------
# syntax       $TaxDigitalChanged = $Basket->updateTaxAreaDigital();
#----------------------------------------------------------------------------------------
# description  Sets the basket tax area digital based on the shipping address country
#               Default tax area digital is the tax area based on the shop address
#
#               Returns true, if calculation of tax has changed, because of changed
#               country in billing or shipping address
#----------------------------------------------------------------------------------------
# return       $TaxDigitalChanged | true if tax area or tax calculation has changed,
#                             otherwise undef | boolean
#========================================================================================
sub updateTaxAreaDigital {
    my $self = shift;
    my $Shop = $self->getSite();
    my $Container = $self->container;
    return unless defined $Container;
    return undef unless $Shop->get('TaxMatrixDigital');

    # try to find the correct taxareadigital for the user
    # order: basket/user/shopfallback
    my $TaxAreaDigital = $self->taxAreaDigitalByAddress();

    my $OldTaxAreaDigital = $Container->get('TaxAreaDigital');
    if (Compare('TaxAreaDigital', $OldTaxAreaDigital, $TaxAreaDigital)) {
        $Container->set({'TaxAreaDigital' => $TaxAreaDigital});
        $self->recalculate;
        return 1;
    }

    return;
}

#========================================================================================
# function     taxAreaDigitalByAddress
# state        public
#----------------------------------------------------------------------------------------
# syntax       $TaxDigital = $Basket->taxAreaDigitalByAddress();
#----------------------------------------------------------------------------------------
# description  Get the basket tax area digital based on the shipping address country
#               Default tax area digital is the tax area based on the shop address
#
#               Returns the tax area digital for the basket, or undef if the address
#
#----------------------------------------------------------------------------------------
# return       $TaxDigital | tax area digital object | object
#========================================================================================
sub taxAreaDigitalByAddress {
    my $self = shift;
    my $Shop = $self->getSite();

    my $TaxAddress;

    return unless $Shop->get('TaxMatrixDigital');

    # BASKET: do we have already a basket address?
    $TaxAddress = $self->get('ShippingAddress') // $self->get('BillingAddress');

    # USER: do we have already a user address?
    $TaxAddress = $self->get('User')->get('ShippingAddress') // $self->get('User')->get('BillingAddress') unless $TaxAddress;

    # SHOP FALLBACK: we use the shop address to get the digital tax matrix
    if (defined $TaxAddress) {
        return undef unless $TaxAddress->get('IsEU');
    }
    else {
        return undef unless $Shop->get('Address')->get('IsEU');
        return $Shop->get('DefaultTaxAreaDigital');
    }

    my $AddressCountryID = $TaxAddress->get('CountryID');
    my $TaxAreaDigital = $Shop->get('TaxMatrixDigital')->taxAreaDigitalByCountry($AddressCountryID);

    return $TaxAreaDigital;
}

#========================================================================================
# function     updatePaymentMethod
# state        public
#----------------------------------------------------------------------------------------
# syntax       $PaymentChanged = $Basket->updatePaymentMethod();
#----------------------------------------------------------------------------------------
# description  Sets the basket payment method based on the shipping/billing address country
#
#               Returns 1 if payment method has changed, because of changed
#               country in billing or shipping address
#----------------------------------------------------------------------------------------
# return       $PaymentChanged | <ul>
#                   <li> 1 : payment method has changed</li>
#                   <li>-1 : no payment method exists | integer</li>
#                   <li>undef : existing payment method is ok</li>
#                   </ul> | integer
#========================================================================================
sub updatePaymentMethod {
    my $self = shift;
    my $LineItemContainer = $self->container;
    return unless defined $LineItemContainer;

    my $PaymentLineItem = $LineItemContainer->get('Payment');
    return -1 unless defined $PaymentLineItem;

    # if the current payment method is no longer valid, select the first valid payment method
    my $aPaymentMethods = $LineItemContainer->get('PaymentMethods');
    return if ElementOf($PaymentLineItem->get('PaymentMethod'), @$aPaymentMethods);

    if (scalar @$aPaymentMethods > 0) {
        SavePayment($LineItemContainer, $aPaymentMethods->[0]);
    }
    else {
        $PaymentLineItem->delete;
    }

    return 1;
}

#========================================================================================
# function     updateShippingMethod
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ShippingChanged = $Basket->updateShippingMethod();
#----------------------------------------------------------------------------------------
# description  Sets the basket shipping method based on the shipping/billing address country
#
#               Returns 1 if shipping method has changed, because of changed
#               country in billing or shipping address
#----------------------------------------------------------------------------------------
# return       $ShippingChanged | <ul>
#                   <li> 1 : shipping method has changed</li>
#                   <li>-1 : no shipping method exists | integer</li>
#                   <li>undef : existing shipping method is ok</li>
#                   </ul> | integer
#========================================================================================
sub updateShippingMethod {
    my $self = shift;
    my $LineItemContainer = $self->container;
    return unless defined $LineItemContainer;

    my $ShippingLineItem = $LineItemContainer->get('Shipping');
    return -1 unless defined $ShippingLineItem;

    # if the current shipping method is no longer valid, select the first valid shipping method
    my $aShippingMethods = $LineItemContainer->get('ShippingMethods');
    return if ElementOf($ShippingLineItem->get('ShippingMethod'), @$aShippingMethods);

    if (scalar @$aShippingMethods > 0) {
        SaveShipping($LineItemContainer, $aShippingMethods->[0]);
    }
    else {
        $ShippingLineItem->delete;
    }

    return 1;

}

sub exportXMLElements {
    my $self = shift;
    my ($Driver, $hAttributes) = @_;

    $self->SUPER::exportXMLElements($Driver, $hAttributes);

    if ($self->class->get('ExportLevel') == $Driver->exportLevel) {
        # add references to child objects
        my $hValues = $self->get([qw(BillingAddress ShippingAddress ShopAddress LineItemContainer)]);
        foreach my $AttributeAlias ( keys %$hValues ) {
            my $Object = delete $hValues->{$AttributeAlias};
            $hValues->{$AttributeAlias} = $self->relativePath($Object) if defined $Object;
        }
        if( scalar keys %$hValues ) {
            $Driver->addObjectReference('Basket', $self, $hValues);
        }
    }
    return;
}

1;
