#========================================================================================
# §package      DE_EPAGES::Product::API::ProductTypeService
# §state        public
#----------------------------------------------------------------------------------------
# §description  Collects the producttypes data of a shop
#========================================================================================
package DE_EPAGES::Product::API::ProductTypeService;

use DE_EPAGES::Object::API::Factory qw ( LoadClassByAlias Transaction );
use DE_EPAGES::WebService::API::Util qw (ReadJSONBoolean);


use strict;



#========================================================================================
# §function     new
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       new( );
#----------------------------------------------------------------------------------------
# §description  Constructor
#----------------------------------------------------------------------------------------
# §return       $self | self | DE_EPAGES::Product::API::ProductTypeService
#========================================================================================
sub new {
    my $class = shift;
    my $self  = bless {}, $class;

    return $self;
}


sub getProductTypes {
    my $self = shift;
    my ($Shop) = @_;

    my $ProductTypesFolder = $Shop->get('ProductTypes');
    return defined $ProductTypesFolder ? $ProductTypesFolder->children : [];

}


sub getProductTypeData {
    my $self = shift;
    my ($ProductType, $LocaleID, $Shop) = @_;

    my $Formatter = DE_EPAGES::Shop::API::Resource::Formatter->new($Shop, $LocaleID);



    my $hProductType = $self->_getProductTypeAttributeValues($ProductType, $LocaleID);

    my $hProductTypeData = {
        'productTypeId' => $hProductType->{'GUID'},
        'alias'         => $hProductType->{'Alias'},
        'name'          => $hProductType->{'NameOrAlias'},
        'attributes'    => $self->_getProductTypeAttributes($ProductType, $LocaleID, $Formatter),
    };

    return $hProductTypeData;
}

sub createProductTypes {
    my $self = shift;
    my($Shop, $LocaleID, $ahRequestBody) = @_;

    my $ProductType;
    my $Locale = $self->_loadLocale($LocaleID);
    my $LanguageID = $Locale->get('LanguageID');
    my @ProductTypes;

    foreach my $NewProductType (@$ahRequestBody) {
        my $ProductTypeAlias = $NewProductType->{'alias'};
        # Trim both ends
        $ProductTypeAlias =~ s/^\s+|\s+$//g;

        my $Formatter = DE_EPAGES::Shop::API::Resource::Formatter->new($Shop, $LocaleID);

        my $ProductTypeFolder = $Shop->folder('ProductTypes');
        my $BaseClass = $Shop->get('ProductClass');
        $self->_Transaction( sub {
            $ProductType = LoadClassByAlias('ProductType')->insertObject( {
                'Alias'   => $self->_getUniqueAliasForProductType($ProductTypeFolder, $NewProductType->{'alias'}),
                'Parent'  => $ProductTypeFolder,
                'Package' => 'DE_EPAGES::Product::API::Object::Product',
                'BaseClass' => $BaseClass,
            } );
        });
        $ProductType->set({'Name' => $NewProductType->{'name'}}, $LanguageID) if defined $NewProductType->{'name'};

        if (defined $NewProductType->{'attributes'}) {
            $self->createProductTypeAttributes($ProductType, $LocaleID, $NewProductType, $Formatter)
        }

        push @ProductTypes , $ProductType;
    }

    return @ProductTypes;
}

sub createProductTypeAttributes {
    my $self = shift;
    my ($ProductType, $LocaleID, $hRequestBody, $Formatter) = @_;
    my $Locale = $self->_loadLocale($LocaleID);
    my $LanguageID = $Locale->get('LanguageID');
    my $Shop = $ProductType->getSite();

    my $AttributesFolder = $ProductType->folder('Attributes');
    my $ahAttributes = $hRequestBody->{'attributes'};

    foreach my $Attribute (@$ahAttributes) {
        if (defined $Attribute->{'name'}) {
            if ($Attribute->{'preDefAttribute'}) {
                _createPreDefinedAttribute($self, $ProductType, $Attribute, $AttributesFolder, $LanguageID);
            } else {
                _createProductAttribute($self, $ProductType, $Attribute, $AttributesFolder, $LanguageID);
            }
        }
    }

    return $self->getProductTypeData($ProductType, $LocaleID, $Shop);
}

sub _getProductTypeAttributeValues {
    my $self = shift;
    my ($ProductType, $LocaleID) = @_;

    my $Locale = $self->_loadLocale($LocaleID);
    my $LanguageID = $Locale->get('LanguageID');

    my @Attributes = qw (
        GUID
        Alias
        NameOrAlias
    );

    return $ProductType->get( \@Attributes, $LanguageID );
}

sub _getProductTypeAttributes {
    my $self = shift;
    my ($ProductType, $LocaleID, $Formatter) = @_;

    my $Locale = $self->_loadLocale($LocaleID);
    my $LanguageID = $Locale->get('LanguageID');

    return $self->_getChildAttributes($ProductType, $LanguageID, $Formatter);
}

sub _getChildAttributes {
    my $self = shift;
    my ($ProductTypeAttribute, $LanguageID, $Formatter) = @_;

    my $aAttributes    = $ProductTypeAttribute->get('ChildObjects');
    my @AttributeValues;


    foreach my $Attribute (@$aAttributes) {
        foreach my $Item (@{$Attribute->get('ChildObjects')}) {
            my $isPreDefAttribute = $Item->get('Package') eq 'DE_EPAGES::Object::API::Attributes::PreDefined' ? 1 : 0;
            push @AttributeValues , {   'name' => $Item->get('Alias'),
                                        'visible' => $Formatter->bool($Item->get('IsVisible')),
                                        'preDefAttribute' => $Formatter->bool($isPreDefAttribute),
                                        'values'  => $isPreDefAttribute ? $self->_getAttributes($Item, $LanguageID) : [],
            };
        }
    }

    return \@AttributeValues;
}

sub _getAttributes {
    my $self = shift;
    my ($ChildAttribute, $LanguageID) = @_;

    my $aAttributes    = $ChildAttribute->get('ChildObjects');
    my @AttributeValues;

    foreach my $Attribute (@$aAttributes) {
        foreach my $Item (@{$Attribute->get('ChildObjects')}) {
            push @AttributeValues, { 'name'     => $Item->get('NameOrAlias', $LanguageID),
                                     'value'    => $Item->get('ValueOrAlias', $LanguageID),
                                    };
        }
    }

    return \@AttributeValues;
}

sub _loadLocale {
    my $self = shift;
    my ($LocaleID) = @_;
    return DE_EPAGES::Object::API::Object::Locale->load($LocaleID);
}

sub _Transaction {
    my $self = shift;
    my ($cCode, $cRollback) = @_;
    Transaction($cCode, $cRollback);
    return;
}

sub _getUniqueAliasForProductType{
    my $self = shift;
    my ($Folder, $aliasMethod) = @_;

    if ($Folder->existsChild($aliasMethod)) {
        my $i = 1;
        my $newAliasMethod = $aliasMethod . '_' .  $i;
        while ($Folder->existsChild($newAliasMethod)) {
            $newAliasMethod =  $aliasMethod . '_' .  ++$i;
        }
        $aliasMethod = $newAliasMethod;
    }
    return $aliasMethod;
}

sub _createPreDefinedAttribute{
    my ($self, $ProductType, $hAttribute, $AttributesFolder, $LanguageID) = @_;

    my $NewAttribute;
    my $Class = LoadClassByAlias('ProductAttribute');
    my $ahAttributeValues = $hAttribute->{'values'};
    $self->_Transaction( sub {
        $NewAttribute = $Class->insertObject( {
            'Alias'          => $self->_getUniqueAliasForProductType($AttributesFolder, $hAttribute->{'name'}),
            'Parent'         => $AttributesFolder,
            'AttributeClass' => $ProductType,
            'Package'        => 'DE_EPAGES::Object::API::Attributes::PreDefined',
            'Type'           => 'PreDefLocalizedString',
            'ExportLevel'    => '1',
            'IsUserDefined'  => '1',
        } );
    });
    $NewAttribute->set({'Name' => $hAttribute->{'name'}}, $LanguageID);
    $NewAttribute->set({'IsVisible' => ReadJSONBoolean($hAttribute->{'visible'})}, $LanguageID) ? defined ReadJSONBoolean($hAttribute->{'visible'}) : 0;

    my $Folder = $NewAttribute->folder('PreDefAttribute');
    my $PreDefAttributeClass = LoadClassByAlias('PreDefAttribute');

    foreach my $Value (@$ahAttributeValues) {
        my $NewAttributeValue;
        if (defined $Value->{'value'}) {
            $self->_Transaction( sub {
                $NewAttributeValue = $PreDefAttributeClass->insertObject( {
                    'Alias'   => $self->_getUniqueAliasForProductType($Folder, $Value->{'name'}),
                    'Parent'  => $Folder,
                });
            });
            $NewAttribute->set({'Name' => $hAttribute->{'name'}}, $LanguageID);
            $NewAttributeValue->set({'ValueLocString' => $Value->{'value'}}, $LanguageID)
        }
    }

    return;
}

sub _createProductAttribute{
    my ($self, $ProductType, $hAttribute, $AttributesFolder, $LanguageID) = @_;

    my $NewAttribute;
    my $Class = LoadClassByAlias('ProductAttribute');
    $self->_Transaction( sub {
        $NewAttribute = $Class->insertObject( {
            'Alias'          => $self->_getUniqueAliasForProductType($AttributesFolder, $hAttribute->{'name'}),
            'Parent'         => $AttributesFolder,
            'AttributeClass' => $ProductType,
            'Package'        => 'DE_EPAGES::Object::API::Attributes::LocalizedAttribute',
            'Type'           => 'LocalizedString',
            'ExportLevel'    => '1',
            'IsUserDefined'  => '1',
        } );
    });
    $NewAttribute->set({'Name' => $hAttribute->{'name'}}, $LanguageID);
    $NewAttribute->set({'IsVisible' => ReadJSONBoolean($hAttribute->{'visible'})}, $LanguageID) ? defined ReadJSONBoolean($hAttribute->{'visible'}) : 0;

    return;
}

1;
