use strict;
use Getopt::Long;
use Time::HiRes qw (time);
use DE_EPAGES::Core::API::Error qw( GetError ExistsError );
use DE_EPAGES::Core::API::Log qw( LogError );
use DE_EPAGES::Core::API::Script qw( RunScript );
use DE_EPAGES::Database::API::Connection qw( RunOnStore GetCurrentDBHandle );
use DE_EPAGES::Object::API::Factory qw ( LoadObject Transaction );
use DE_EPAGES::Shop::API::System qw( GetUnityShopIDs );
use DE_EPAGES::WebInterface::API::MessageCenter qw( SynchronizeCache );
use DE_EPAGES::Object::API::Table::Language qw( GetAllInfoLanguages );
use DE_EPAGES::Core::API::PerlTools qw ( isValid );

#========================================================================================
# §function         ImportCustomizableHeadlineDefaultText
# §state            public
#----------------------------------------------------------------------------------------
# §syntax           ImportCustomizableHeadlineDefaultText();
#----------------------------------------------------------------------------------------
# §description      Imports the CustomizableHeadlineDefaultText translations in all
#                   shop languages
#========================================================================================
sub ImportCustomizableHeadlineDefaultText {

    my $dbi = GetCurrentDBHandle();
    if($dbi->driverName() eq 'mysql'){
        my $hStructure = GetTranslations();

        print "Reading all Unity shops\n";
        my $aShopIDs = GetUnityShopIDs();

        foreach my $ShopID (@$aShopIDs) {
            my $Shop = LoadObject($ShopID);
            PatchAttributes($hStructure, $Shop)
        }

        eval { SynchronizeCache(); };
        LogError(GetError()) if ExistsError();

        return;
    }
}

sub PatchAttributes {
    my ($hStructure, $Shop) = @_;

    my $dbi = GetCurrentDBHandle();
    my $Attributename = 'CustomizableProductHeadlineDefaultText';
    my $AttributeID   = $Shop->class->attribute('CustomizableProductHeadlineDefaultText')->id;
    my $ObjectID   = $Shop->id;

    # get all Languages
    my $ahLanguages = [ grep {$_->{'IsUsed'}} @{GetAllInfoLanguages()} ];

    # start transaction for all attribute changes
    Transaction(sub {
        my $sql;
        foreach my $hLanguage (@$ahLanguages) {

            my $LanguageCode = $hLanguage->{'Code2'};
            my $LanguageID = $hLanguage->{'LanguageID'};
            my $Dialect = $hLanguage->{'Dialect'};

            $LanguageCode .= "-$Dialect" if $Dialect;

            next unless exists $hStructure->{$LanguageCode};


            # get value from structure
            my $AttributeValue = $hStructure->{$LanguageCode}->{$Attributename};
            die "Missing value for $Attributename/$LanguageCode" unless $AttributeValue;
            my $LocalizedString = $Shop->get('CustomizableProductHeadlineDefaultText', $LanguageID);

            unless(isValid($LocalizedString)) {
                $sql = <<'END_SQL';
INSERT INTO localizedstringattribute (objectid, attributeid, languageid, value)
VALUES (?, ?, ?, ?)
END_SQL
                $dbi->do($sql, [
                    $ObjectID,
                    $AttributeID,
                    $LanguageID,
                    $AttributeValue
                ]);
            }
        }# end foreach Language
    });
}



sub GetTranslations {
    return {
        'en' => {
            'CustomizableProductHeadlineDefaultText' => 'Add customization',
        },
        'de' => {
            'CustomizableProductHeadlineDefaultText' => 'Personalisierungswunsch',
        },
        'ca' => {
            'CustomizableProductHeadlineDefaultText' => 'Afegeix personalització',
        },
        'cs' => {
            'CustomizableProductHeadlineDefaultText' => 'Přidat přizpůsobení',
        },
        'da' => {
            'CustomizableProductHeadlineDefaultText' => 'Tilføj tilpasning',
        },
        'es' => {
            'CustomizableProductHeadlineDefaultText' => 'Añadir personalización',
        },
        'fi' => {
            'CustomizableProductHeadlineDefaultText' => 'Lisää tuotteen mukautus',
        },
        'fr' => {
            'CustomizableProductHeadlineDefaultText' => 'Personnaliser',
        },
        'it' => {
            'CustomizableProductHeadlineDefaultText' => 'Aggiungi testo personalizzato',
        },
        'nl' => {
            'CustomizableProductHeadlineDefaultText' => 'Aanpassing toevoegen',
        },
        'no' => {
            'CustomizableProductHeadlineDefaultText' => 'Legg til tilpasning',
        },
        'ru' => {
            'CustomizableProductHeadlineDefaultText' => 'Add customisation',
        },
        'sv' => {
            'CustomizableProductHeadlineDefaultText' => 'Anpassa',
        },
        'en-US' => {
            'CustomizableProductHeadlineDefaultText' => 'Add customization',
        },
        'pt' => {
            'CustomizableProductHeadlineDefaultText' => 'Adicionar personalização',
        },
        'pt-BR' => {
            'CustomizableProductHeadlineDefaultText' => 'Adicionar personalização',
        },
        'es-MX' => {
            'CustomizableProductHeadlineDefaultText' => 'Agregar personalización',
        },
    };
}

sub Main {

    local $| = 1;

    my $Password;
    my $StoreName;
    my $Help;

    GetOptions(
        'help'        => \$Help,
        'passwd=s'    => \$Password,
        'storename=s' => \$StoreName,
    );

    usage() if $Help;

    unless ( defined $StoreName ) {
        print STDERR "Missing parameter -storename.\n\n";
        usage();
    }
    my $time1 = time;

    RunOnStore(
        Store      => $StoreName,
        DBPassword => $Password,
        Sub        => sub { ImportCustomizableHeadlineDefaultText(); },
    );
    my $time2 = time;
    printf "need %.1f seconds\n", $time2 - $time1;
}

sub usage {
    print <<END_USAGE;
Description:
    Imports the CustomizableHeadlineDefaultText translations in all shop languages


Usage:
    perl $0 [options] [flags] datafiles

options:
    -passwd     database user password                          ()
    -storename  name of store                                   ()
flags:
    -help       show the command line options

Example:
    perl $0 -storename Store
END_USAGE
    exit 2;
}

RunScript( 'Sub' => \&Main );
